#####################################################################
# Copyright (C) 2000 Artur Trzewik
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
######################################################################
# Header
# CVS - $Id: widgetlib.tcl,v 1.12 2000/10/15 17:19:20 artur Exp $
#
# Information:
# Modul (benutzt von Schema Editor)
# Bibliothek in XOTCL fr Tk und Tix Widget und Windget Container
# Kappselung von Tix/Tk Widget als XOTcl Klassen
Class OWidget -parameter {win}
#OWidget abstract instproc map window
OWidget instproc unmap {} {
    [self] instvar win
    destroy $win
}
Class InputWidget -superclass OWidget -parameter {
    value
    data
    {default {}}
    {state normal}
}
InputWidget metadata add description 
InputWidget metadata description {
Eine abstrakte Klasse fr alle InputWidget
Parameter: value default state
Methoden: getvalue, setdefault, setstatus status
}
# die Methoden getvalue, setvalue, setdefault mssen nich beschrieben werden
# wenn value per OPtion -varible gelinkt ist.
# Das immer den InputWidget Zustand widerspiegelt
InputWidget instproc getvalue {} {
    [self] instvar value
    return $value
}
InputWidget instproc setdefault {} {
    [self] instvar default value
    set value $default
}
InputWidget instproc setvalue newv {
    [self] instvar value
    set value $newv
}
InputWidget instproc setstatus {astatus} {
    [self] instvar win status
    set status $astatus
    $win configure -status $astatus
}
InputWidget instproc updatevalue {} {
}

Class OLabelEntry -superclass InputWidget -parameter {
    {label {}}
    show
    textwidth
}
OLabelEntry metadata add description
OLabelEntry metadata description {
Kappselung der tixLabelEntry
Parameter: label show
}
OLabelEntry metadata add attributes
OLabelEntry metadata attributes {state label show textwidth}

OLabelEntry instproc map window {
    [self] instvar label win state show textwidth
    #set value $default
    set xparam {}
    set subparam {}
    if {[info exist show]} {
	lappend subparam entry.show *
    }
    if {[info exist textwidth]} {
	lappend subparam entry.width $textwidth
    }
    if {$subparam!=""} {
	lappend xparam -options $subparam
    }
    eval tixLabelEntry $window -label [list $label] -state [list $state] $xparam
    set win $window 
    [self] setdefault
    next
}
OLabelEntry instproc getvalue {} {
    [self] instvar value win
    [self] updatevalue
    return $value
}
OLabelEntry instproc updatevalue {} {
    [self] instvar value win
    set value [[$win subwidget entry] get]
}
OLabelEntry instproc setvalue {newv} {
    [self] instvar win value state
    set value $newv
    set ewin [$win subwidget entry]
    # man kann in den edit widget bei disabled nicht schreiben (seltsam)
    if {$state=="disabled"} {
	$ewin configure -state normal
    }
    $ewin delete 0 end
    $ewin insert 0 $value
    if {$state=="disabled"} {
	$ewin configure -state disabled
    }
}
OLabelEntry instproc setdefault {} {
    [self] instvar default
    [self] setvalue $default
}
Class OLabelText -superclass InputWidget -parameter {
    {label {}}
    {width 40}
    {height 8}
}
OLabelText metadata add description
OLabelText metadata description {
tixScrolledText im tixLabelFrame
Parameter: label state label width height
}
OLabelText metadata add attributes
OLabelText metadata attributes {state label width height}

OLabelText instproc map window {
    [self] instvar label win state width height
    #set value $default
    set xparam {}
    set subparam {}
    if {[info exist width]} {
	lappend subparam text.width $width
    }
    if {[info exist height]} {
	lappend subparam text.height $height
    }
    if {[info exist state]} {
	lappend subparam text.state $state
    }
    if {$subparam!=""} {
	lappend xparam -options $subparam
    }
    tixLabelFrame $window -label [list $label]
    eval tixScrolledText $window.stext $xparam -scrollbar auto
    pack $window.stext -expand yes -fill both -in [$window subwidget frame]
    set win $window 
    [self] setdefault
    next
}
OLabelText instproc getvalue {} {
    [self] instvar value win
    [self] updatevalue
    return $value
}
OLabelText instproc updatevalue {} {
    [self] instvar value win
    set ewin [$win.stext subwidget text]
    # Leeres Text Widget gibt immer \n zurck (seltsam aber wahr) darum trim
    set value [string trim [$ewin get 1.0 end]]
}
OLabelText instproc setvalue {newv} {
    [self] instvar win value
    set value $newv
    set ewin [$win.stext subwidget text]
    $ewin delete 1.0 end
    $ewin insert 1.0 $value
}
OLabelText instproc setdefault {} {
    [self] instvar default
    [self] setvalue $default
}
Class OEditList -superclass InputWidget -parameter {
    {label {}}
}
OEditList metadata add description
OEditList metadata description {
tixScrolledText im tixLabelFrame
Parameter: label state
}
OEditList metadata add attributes
OEditList metadata attributes {state label width height}

OEditList instproc map window {
    [self] instvar label value default win state textwidth
    tixComboBox $window -label $label -state $state \
	-editable true -history true -dropdown false
    set win $window
    button $window.del -text Del -command "[self] deleteItem"
    # Vorsicht Tix Interna werden hier berhrt (dirty fix)
    pack $window.del -side right -in $window.frame.f1 \
	-before [$window subwidget entry]
    [self] setdefault
    next
}
OEditList instproc deleteItem {} {
    [self] instvar win
    set lwin [$win subwidget listbox]
    set sel [$lwin curselection]
    if {$sel!=""} {$lwin delete $sel}
}
OEditList instproc getvalue {} {
    [self] instvar value win
    [self] updatevalue
    return $value
}
OEditList instproc updatevalue {} {
    [self] instvar value win
    set lwin [$win subwidget listbox]
    set value [$lwin get 0 end]
}
OEditList instproc setvalue {newv} {
    [self] instvar win value
    set value $newv
    [$win subwidget listbox] delete 0 end
    foreach elem $value {
	$win insert end $elem
    }
}
OEditList instproc setdefault {} {
    [self] instvar default
    [self] setvalue $default
}

Class OLabelControl -superclass InputWidget -parameter {
    {label {}}
    min
    max
}
OLabelControl metadata add description
OLabelControl metadata description {
Kappselung der tixControl
Benutzt um Integer Input zu realisieren
Parameter: label min max
}
OLabelControl metadata add attributes
OLabelControl metadata attributes {state label min max}
OLabelControl instproc map window {
    [self] instvar label value default win state min max
    if {$default=="NULL"} {set default ""}
    set value $default
    set xparam {}
    if {[info exist max]} {
	lappend xparam -max $max
    }
    if {[info exist min]} {
	lappend xparam -min $min
    }
    # Leider funktioniert  die Option allowempty nicht
    # Bug in Control.tix
    # deswegen eigene -validatecmd
    eval tixControl $window -label [list $label] -state $state -variable [self]::value $xparam -integer 1 -allowempty 1 -validatecmd \"[self] validatecmd\"
    
    set win $window 
    next
}
# Die Orginalle tixControl::SetValue ist fehlerhaft
# es bercksichtigt kein allowempty
# Deswegen eigene Prozedur
OLabelControl instproc validatecmd newvalue {
    [self] instvar min max
    if {$newvalue!="" && [catch {expr 0+$newvalue}]} {
	set newvalue 0
	set data(-value) 0
	set changed 1
    }
    
    if {$newvalue == ""} {
	set ret ""
    }
    
    if {$newvalue != ""} {
	# Change this to a valid decimal string (trim leading 0)
	#
	regsub {^[0]*} $newvalue "" newvalue
	if [catch {expr 0+$newvalue}] {
	    set newvalue 0
	    set ret 0
	    set changed 1
	}
	
	set ret $newvalue

	# Now perform boundary checking
	#
	if {[info exist max] && $max != "" && $ret > $max} {
	    set ret $max
	}
	if {[info exist min] && $min != "" && $ret < $min} {
	    set ret $min
	}
    }
    return $ret
}
# Die nchsten zwei Methoden brauchen nur eine berschreibung weil
# bei state disable kein Wert geschreiben werden konnte
OLabelControl instproc setdefault {} {
    [self] instvar default
    if {$default=="NULL"} { set default "" }
    [self] setvalue $default
}
OLabelControl instproc setvalue {newv} {
    [self] instvar win value state
    # man kann in den edit widget bei disabled nicht schreiben (seltsam)
    if {$state=="disabled"} {
	$win configure -state normal
    }
    set value $newv
    if {$state=="disabled"} {
	$win configure -state disabled
    }
}
# Hack um besonderen Wert "NULL" 
OLabelControl instproc getvalue {} {
    [self] instvar value
    if {$value==""} {
	return NULL
    } else {
	return $value
    }
}
OLabelControl instproc updatevalue {} {
    [self] instvar win
    if {[winfo exist $win]} {
	$win update
    }
}

Class OCheckButton -superclass InputWidget -parameter {
    label
    {default 0}
    {onvalue 1}
    {offvalue 0}
}
OCheckButton metadata add description 
OCheckButton metadata description {
Kappselung des checkbuttons von Tk
Parameter: label
}
OCheckButton metadata add attributes
OCheckButton metadata attributes {state label onvalue offvalue}

OCheckButton instproc map window {
    [self] instvar label value default win state onvalue offvalue 
    set value $default
    checkbutton $window -text $label -state $state -variable [self]::value \
	-onvalue $onvalue -offvalue $offvalue
    set win $window 
    next
}

Class OListChooser -superclass InputWidget -parameter {
    valuelist
}
Class OComboBox -superclass OListChooser  -parameter {label {editable 0}}
OComboBox metadata add attributes
OComboBox metadata attributes {state label valuelist}
OComboBox instproc map window {
    [self] instvar label value default win state valuelist editable
    set value $default
    if {$value==""} {
	set value [lindex $valuelist 0]
    }
    tixComboBox $window -label $label -state $state -variable [self]::value \
	-editable $editable -dropdown true
    foreach elem $valuelist {
	$window insert end $elem
    }
    set win $window
    next
}
OComboBox instproc setvaluelist list {
    [self] instvar win valuelist
    set valuelist $list
    set lwin [$win subwidget listbox]
    $lwin delete 0 end
    foreach elem $valuelist {
	$win insert end $elem
    }
}
# Eine Erweiterung benutzt um eine ChangeValue Notify Handler zu schaffen
Class OComboBoxN -superclass OComboBox -parameter {{notifyproc {}}}
OComboBoxN instproc map args {
    next
    [self] instvar win notifyproc
    $win configure -command $notifyproc 
    #puts "Configured [self] $notifyproc"
}
Class ORadioButtons -superclass OListChooser -parameter {
    {label {}}
    {allowzero 1}
}
ORadioButtons metadata add description
ORadioButtons metadata description {
Abgeleitet from tixSelect
Parameter:  state valuelist
}
ORadioButtons metadata add attributes
ORadioButtons metadata attributes {state label valuelist}
ORadioButtons instproc map window {
    [self] instvar label win state valuelist allowzero 
    tixSelect $window -state $state -label $label -radio true \
	-allowzero $allowzero -orientation vertical
    set index 0 
    foreach elem $valuelist {
	$window add but_$index -text $elem
	incr index
    }
    set win $window
    [self] setdefault
    next
}
ORadioButtons instproc getvalue {} {
    [self] instvar value win
    [self] updatevalue
    return $value
}
ORadioButtons instproc updatevalue {} {
    [self] instvar value valuelist win
    regexp {but_([0-9]+)} [$win cget -value] _ index
    if {[info exist index]} {
	set value [lindex $valuelist $index]
    }
}
ORadioButtons instproc setvalue {newv} {
    [self] instvar win value valuelist
    set value $newv
    set index  [lsearch $valuelist $value]
    if {$index==-1} {
	$win configure -value {} 
    } else {
	if {[$win cget -value]!="but_$index"} {
	    $win invoke but_$index
	}
    }
}
ORadioButtons instproc setdefault {} {
    [self] instvar win default
    [self] setvalue $default
}
Class OMultiList -superclass OListChooser -parameter {label}
OMultiList metadata add description
OMultiList metadata description {
Abgeleitet from tixScrolledList ermglicht Auswahl von Untermengen
Parameter:  state valuelist
}
OMultiList metadata add attributes
OMultiList metadata attributes {state valuelist}
OMultiList instproc map window {
    [self] instvar win state valuelist
    tixScrolledListBox $window -state $state -scrollbar y \
	-options { listbox.selectMode multiple }
    set win $window
    [self] setvaluelist $valuelist
    [self] setdefault
    next
}
OMultiList instproc getvalue {} {
    [self] instvar value win
    [self] updatevalue
    return $value
}
OMultiList instproc updatevalue {} {
    [self] instvar value valuelist win
    set lwin [$win subwidget listbox]
    set value {}
    foreach index [$lwin curselection] {
	lappend value [lindex $valuelist $index]
    }
    #puts "ML value $value"
}
OMultiList instproc setvalue {newv} {
    [self] instvar win value valuelist
    set value $newv
    set lwin [$win subwidget listbox]
    $lwin selection clear 0 end
    foreach elem $value {
	$lwin selection set [lsearch $valuelist $elem]
    }
}
OMultiList instproc setdefault {} {
    [self] instvar default
    [self] setvalue $default
}
OMultiList instproc setvaluelist list {
    [self] instvar win valuelist
    set valuelist $list
    set lwin [$win subwidget listbox]
    $lwin delete 0 end
    foreach elem $valuelist {
	$lwin insert end $elem
    }
}
# Dieses Widget wird verwendet um Auswahllisten
# mit mehreren Splaten zu realisieren
# fr Zwecke der Anzeige von humankonformen Fremdschlssel
Class OMultiColumnList  -superclass OWidget -parameter {
    label 
    columns 
    {columnslabel {}}
    {height 12}
    {columnswidth 12}
    {state normal}
    {dropdown 1}
} 
OMultiColumnList metadata add description
OMultiColumnList metadata description {
Mehrspaltige Auswahlliste
hat kein Attribut valuelist
setvalue und getvalue behandeln den Index statt konkreten Werten.
Hat kein setdefault value
Parameter:  state columns columnslabel columnswidth height
}
OMultiColumnList metadata add attributes
OMultiColumnList metadata attributes {state label columns columnslabel height columnswidth}
OMultiColumnList instproc map window {
    [self] instvar win state label columns columnslabel columnswidth height dropdown lb_root
    set win $window
    tixLabelFrame $window -label $label
    bind $win <FocusOut> "[self] popdown"
    # zu kompliziertes Algorith ausgliedern !!
    if {$dropdown} {
	menu $win.dropshell -bd 2 -relief raised -tearoff 0
	wm overrideredirect $win.dropshell 1
	wm withdraw $win.dropshell
	set lb_root $win.dropshell
    } else {
	set lb_root $win
    }
    for {set x 0} {$x<$columns} {incr x} {
	set width [lindex $columnswidth $x]
	if {$width==""}  {set width [lindex $columnswidth 0]}
	set elem [listbox $lb_root.listbox$x -height $height -width $width \
		      -yscrollcommand "$lb_root.scroll set" \
		      -exportselection no]
	bind $elem <1> "[self] setindex \[%W index @%x,%y\]"
	bind $elem <Key-Select> "[self] setindex \[%W index @%x,%y\]"
	if {$dropdown} {
	    set elem [::label $win.labentry$x -borderwidth 1 -width $width -justify left -relief solid]
	    bind $elem <1> "[self] popup"
	    pack $lb_root.listbox$x -expand yes \
		-fill both -side left
	}
	if {$columnslabel!=""} {
	    ::label $window.label$x -text  [lindex $columnslabel $x]
	    grid $window.label$x -column $x -row 0 -sticky nsew  -in [$window subwidget frame]
	    grid $elem -column $x -row 1 -sticky ns  -in [$window subwidget frame]
	} else {
	    pack $elem -expand yes \
		-fill both -side left -in [$window subwidget frame]
	}
    }
    scrollbar $lb_root.scroll -command "[self] yview"
    if {$dropdown} {
	pack $win.dropshell.scroll -expand yes \
		-fill y -side left
    } else {
	if {$columnslabel!=""} {
	    grid $window.scroll -column $x -row 1 -sticky ns  -in [$window subwidget frame]
	} else {
	    pack $window.scroll -expand yes \
		-fill y -side left -in [$window subwidget frame]
	}
    }
    next
}
OMultiColumnList instproc yview {args} {
    [self] instvar lb_root columns
    for {set x 0} {$x<$columns} {incr x} {
	eval $lb_root.listbox$x yview $args
    }

}
OMultiColumnList instproc popup {args} {
    [self] instvar win lb_root
    #puts "popup"
    if {![winfo ismapped $win]} {
        return
    }
    set  y [winfo rooty $win]
    incr y [winfo height $win]

    set height [winfo reqheight $lb_root.listbox0]

    set x1 [winfo rootx $win]
    set reqwidth [winfo reqwidth $win.dropshell]
    set width $reqwidth
    set x2 $width

    # If the listbox is too far right, pull it back to the left
    #
    set scrwidth [winfo vrootwidth .]
    if {$x2 > $scrwidth} {
        set x1 [expr $scrwidth - $width]
    }

    # If the listbox is too far left, pull it back to the right
    #
    if {$x1 < 0} {
        set x1 0
    }

    # If the listbox is below bottom of screen, put it upwards
    #
    # If the listbox is below bottom of screen, put it upwards
    #
    set scrheight [winfo vrootheight .]
    set bottom [expr $y+$height]
    if {$bottom > $scrheight} {
        set y [expr $y-$height-[winfo height $win]-5]
    }

    # OK , popup the shell
    #
    global tcl_platform

    wm geometry $win.dropshell $reqwidth\x$height+$x1+$y
    if {$tcl_platform(platform) == "windows"} {
        update
    }
    wm deiconify $win.dropshell
    if {$tcl_platform(platform) == "windows"} {
        update
    }
    raise $win.dropshell
    focus $win
    #puts "end of popup"
}
OMultiColumnList instproc popdown {args} {
    [self] instvar win
    wm withdraw $win.dropshell 
}
OMultiColumnList instproc getindex {} {
    [self] instvar lb_root win
    return [$lb_root.listbox0 curselection]
}
OMultiColumnList instproc setindex {index} {
    [self] instvar lb_root columns dropdown win
    #puts "set index to $index"
    for {set x 0} {$x<$columns} {incr x} {
	$lb_root.listbox$x selection clear 0 end
	if {$index!=""} { 
	    $lb_root.listbox$x selection set $index
	    if {$dropdown} {
		$win.labentry$x configure -text [$lb_root.listbox$x get $index]
	    }
	} elseif {$dropdown} {
	    $win.labentry$x configure -text ""
	}
    }
    if {$dropdown} {
	[self] popdown
    }
}
OMultiColumnList instproc setvaluelist list {
    [self] instvar lb_root columns
    for {set x 0} {$x<$columns} {incr x} {
	$lb_root.listbox$x delete 0 end
    }
    foreach elem $list {
	for {set x 0} {$x<$columns} {incr x} {
	    $lb_root.listbox$x insert end [lindex $elem $x]
	}
    }
}
Class OSqlReferenceList  -superclass {InputWidget OMultiColumnList} -parameter {
    sqlquery
    sqlconnect
    {notnull 0}
    {showfirst 1}
    value
    {default {}}
    data
} 
OSqlReferenceList metadata add description
OSqlReferenceList metadata description {
Humankonforme Representation der Fremdschlssel als 
mehrspaltige Auswahlliste
hat kein Attribut valuelist
setvalue und getvalue behandeln den Index statt konkreten Werten.
Hat kein setdefault value
Vorsicht Parameter sqlconnect mu initialisiert werden
auch notnull gehrt nicht zur Klassenvarible attributes
(Stadart Datenverkunft Parameter)
Parameter:  state valuelist columns columnslabel columnswidth height notnull
sqlconnect sqlquery
}
OSqlReferenceList metadata add attributes
OSqlReferenceList metadata attributes {state label columns columnslabel height columnswidth sqlquery showfirst dropdown}
OSqlReferenceList instproc map window {
    #[self] instvar    
    next
    refresh
}

OSqlReferenceList instproc refresh {} {
    [self] instvar sqlconnect sqlquery notnull columns indexlist showfirst 
    if {$sqlconnect==""} return
    set query [$sqlconnect query $sqlquery]
    if {!$notnull} {
	for {set x 0} {$x<$columns} {incr x} {
	    lappend nulllist NULL
	}
	lappend valuelist $nulllist
	lappend indexlist NULL
    }
    for {set x [$query rows]} {$x>0} {incr x -1} {
	# is der Eigentliche Fremschlssel ausgeblendet
	set row  [$query fetch]
	lappend indexlist [lindex $row 0]
	if {$showfirst} {
	    lappend valuelist $row
	} else {
	    lappend valuelist [lrange $row 1 end]
	}
    }
    $query destroy
    [self] setvaluelist $valuelist
}
OSqlReferenceList instproc getvalue {} {
    [self] instvar value win
    [self] updatevalue
    return $value
}
OSqlReferenceList instproc updatevalue {} {
    [self] instvar value win indexlist 
    set index [[self] getindex]
    if {$index!=""} {
	set value [lindex $indexlist $index]
    }
    #puts "upadet sqllist $value"
}
OSqlReferenceList instproc setvalue {newv} {
    [self] instvar win value indexlist
    global lan
    set value $newv
    # Lerres Wert (z.B. neuer Tulep 
    if {$value==""} {
	[self] setindex {}
	return
    }
    set index [lsearch $indexlist $value]
    if {$index==-1} {
	tk_messageBox -message $lan(dialog.noprimarykey) -type ok -icon warning
	return
    }
    [self] setindex $index
    [self] yview $index
}
OSqlReferenceList instproc setdefault {} {
    [self] instvar win default
    [self] setvalue $default
}
# Ein Widget fr Eingabe von Array Strukturen 
# dabei wird ein Frame als Eigabe Maske benutzt
# weiter werden Navigations Elemente fr Blttern durch das Array gebildet
# Variable value wird wie folg gebildet
# keyed list (tclX)
# {{name {val1 val2 val3}} {name2 {val1 val2 val3}}}
# array von 3 Elementen
Class OMultiFrame -superclass InputWidget -parameter {
    {usernew 1}
    {userdelete 1}
    {count 1}
    inputframe
}
OMultiFrame instproc map {window} {
    [self] instvar win inputframe usernew userdelete actuel default count
    set win $window
    set actuel 0
    frame $win -bd 2 -relief sunken
    set commands {
	{prev back prev}
	{next forward next_tupel}
	{new filenew new}
	{delete delete delete}
    }
    frame $win.navigation -bd 2 -relief sunken
    label $win.navigation.pos
    pack $win.navigation.pos -side left -anchor w
    foreach command $commands {
	set name [lindex $command 0]
	if {!$usernew && $name=="new"} continue
	if {!$userdelete && $name=="delete"} continue
	set image [lindex $command 1]
	set execute [lindex $command 2]
	button $win.navigation.$name -image [tix getimage $image] \
	    -command "[self] $execute" \
	    -takefocus 0 -bg #c0c0c0 -highlightbackground #c0c0c0
	pack $win.navigation.$name -side left -anchor w
    }
    $inputframe map $win.inputframe
    pack $win.inputframe
    pack $win.navigation -side bottom -anchor s -expand yes -fill x    
    if {$default==""} {
	set default [$inputframe get_values]
    } else {
	foreach key [keylkeys default] {
	    if {$count<[llength [keylget default $key]]} {
		set count [llength [keylget default $key]]
	    }
	}
    }
    [self] setdefault
    next
}
OMultiFrame instproc getvalue {} {
    [self] instvar value win
    [self] updatevalue
    return $value
}
OMultiFrame instproc updatevalue {} {
    [self] updateindex
}
OMultiFrame instproc setvalue {newv} {
    [self] instvar win value actuel
    set value $newv
    [self] showindex $actuel
}
OMultiFrame instproc setdefault {} {
    [self] instvar default
    [self] setvalue $default
}
OMultiFrame instproc showindex {id} {
    [self] instvar value actuel inputframe win count
    set actuel $id
    set indexval {}
    foreach key [keylkeys value] {
	keylset indexval $key [lindex [keylget value $key] $id]
    }
    $inputframe set_defaults $indexval
    $win.navigation.pos configure -text "$actuel von $count"
}
OMultiFrame instproc updateindex {} {
    [self] instvar value actuel inputframe
    $inputframe update_values
    set changes [$inputframe get_changes]
    foreach key [keylkeys changes] {
	set vallist [keylget value $key]
	if {[llength $vallist]<=$actuel} {
	    for {set x [llength $vallist]} {$x<=$actuel} {incr x} {
		lappend vallist {}
	    }
	}
	keylset value $key [lreplace $vallist $actuel $actuel [keylget changes $key]]
    }
}
OMultiFrame instproc next_tupel {} {
    [self] instvar actuel count
    set id $actuel
    incr id
    if {$id<$count} {
	[self] updateindex
	[self] showindex $id
    }
}
OMultiFrame instproc prev {} {
    [self] instvar actuel
    set id $actuel
    incr id -1
    if {$id>=0} {
	[self] updateindex
	[self] showindex $id
    }
}
OMultiFrame instproc new {} {
    [self] instvar count
    incr count
    [self] showindex [expr $count -1]
}
OMultiFrame instproc delete {} {
    [self] instvar actuel value count
    if {$count>1} {
	incr count -1
	incr actuel -1
	if {$actuel<0} {
	    set actuel 0
	}
	foreach key [keylkeys value] {
	    keylset value $key [lreplace [keylget value $key] $actuel $actuel]
	}
	[self] showindex $actuel
    }
}
# Vereinfacht die Manipulation von value keyed list
OMultiFrame instproc getidvalue {id} {
    [self] instvar value 
    set indexval {}
    foreach key [keylkeys value] {
	keylset indexval $key [lindex [keylget value $key] $id]
    }
    return $indexval
}
Class WidgetContainer -superclass OWidget
WidgetContainer instproc get_values {} {
    set inputs [FrameVisitor [self]::inputsvisitor]
    [self] foreach_elem $inputs get_value
    set ret [$inputs set valuelist]
    $inputs destroy
    return $ret
}
# Gibt alle Werte bei dennen dafault und value unterschiedlich sind
# (also wurden sie durch Benutzer verndert)
WidgetContainer instproc get_changes {} {
    set inputs [FrameVisitor [self]::inputsvisitor]
    [self] foreach_elem $inputs get_change
    set ret [$inputs set valuelist]
    $inputs destroy
    return $ret
}
# Setzt Standartwerte
WidgetContainer instproc set_defaults {keyedlist} {
    set inputs [FrameVisitor [self]::inputsvisitor -valuelist $keyedlist]
    $inputs set keys [keylkeys keyedlist]
    [self] foreach_elem $inputs set_default
    $inputs destroy
}
# Setzt Werte (wenn sie anders als Default sind) werden sie
# von get_changes erfasst
WidgetContainer instproc set_values {keyedlist} {
    set inputs [FrameVisitor [self]::inputsvisitor -valuelist $keyedlist]
    $inputs set keys [keylkeys keyedlist]
    [self] foreach_elem $inputs set_value
    $inputs destroy
}
# Muss vor jedem get_values oder get_changes durchgefhrt werden
WidgetContainer instproc update_values {} {
    set uv [FrameVisitor [self]::updatevisitor]
    [self] foreach_elem $uv update_value
    $uv destroy
}
# Setzt defaults auf aktuelle Werten
WidgetContainer instproc update_defaults {} {
    set uv [FrameVisitor [self]::updatevisitor]
    [self] foreach_elem $uv update_default
    $uv destroy
}
# Setzt Werte auf Defaults
WidgetContainer instproc undo_changes {} {
    set uv [FrameVisitor [self]::updatevisitor]
    [self] foreach_elem $uv undo_change
    $uv destroy
}

# Hier wre vielleicht besser ein allg. Composite Objekt
Class FrameVisitor -parameter {
    {valuelist {}}
}
FrameVisitor instproc get_value obj {
    [self] instvar valuelist 
    set objsclass [[$obj info class] info heritage]
    #puts "getvalue $obj $objsclass"
    if {[lsearch $objsclass ::InputWidget]>=0} then {
	if {[$obj info vars data]!=""} {
	    set name [$obj set data]
	} else {
	    regexp {.*::(.+)} $obj _ name
	}
	keylset valuelist $name [$obj set value]
	#lappend valuelist [list $name [$obj set value]]
	#puts "value $valuelist"
    } elseif {[lsearch $objsclass ::WidgetContainer]>=0} {
	$obj foreach_elem [self] get_value
    }
}
FrameVisitor instproc set_default obj {
    [self] instvar valuelist keys
    set objsclass [[$obj info class] info heritage]
    #puts "getvalue $obj $objsclass"
    if {[lsearch $objsclass ::InputWidget]>=0} then {
	if {[$obj info vars data]!=""} {
	    set name [$obj set data]
	} else {
	    regexp {.*::(.+)} $obj _ name
	}
	if {[lcontain $keys $name]} {
	    $obj set default [keylget valuelist $name]
	    if {[$obj info vars win]!=""} {
		$obj setdefault
	    }
	}
	#lappend valuelist [list $name [$obj set value]]
	#puts "value $valuelist"
    } elseif {[lsearch $objsclass ::WidgetContainer]>=0} {
	$obj foreach_elem [self] set_default
    }
}
FrameVisitor instproc set_value obj {
    [self] instvar valuelist keys
    set objsclass [[$obj info class] info heritage]
    #puts "getvalue $obj $objsclass"
    if {[lsearch $objsclass ::InputWidget]>=0} then {
	if {[$obj info vars data]!=""} {
	    set name [$obj set data]
	} else {
	    regexp {.*::(.+)} $obj _ name
	}
	if {[lcontain $keys $name]} {
	    $obj setvalue [keylget valuelist $name]
	}
	#lappend valuelist [list $name [$obj set value]]
	#puts "value $valuelist"
    } elseif {[lsearch $objsclass ::WidgetContainer]>=0} {
	$obj foreach_elem [self] set_value
    }
}
FrameVisitor instproc get_change obj {
    [self] instvar valuelist 
    set objsclass [[$obj info class] info heritage]
    #puts "getvalue $obj $objsclass"
    if {[lsearch $objsclass ::InputWidget]>=0} then {
	if {[$obj info vars data]!=""} {
	    set name [$obj set data]
	} else {
	    regexp {.*::(.+)} $obj _ name
	}
	if {[$obj set value]!=[$obj set default]} {
	    keylset valuelist $name [$obj set value]
	    #lappend valuelist [list $name [$obj set value]]
	}
	#puts "value $valuelist"
    } elseif {[lsearch $objsclass ::WidgetContainer]>=0} {
	$obj foreach_elem [self] get_change
    }
}
FrameVisitor instproc update_value obj {
    set objsclass [[$obj info class] info heritage]
    #puts "update $obj $objsclass"
    if {[lsearch $objsclass ::InputWidget]>=0} then {	
	$obj updatevalue
    } elseif {[lsearch $objsclass ::WidgetContainer]>=0} {
	$obj foreach_elem [self] update_value
    }
}
FrameVisitor instproc update_default obj {
    set objsclass [[$obj info class] info heritage]
    #puts "update $obj $objsclass"
    if {[lsearch $objsclass ::InputWidget]>=0} then {	
	$obj set default [$obj set value]
    } elseif {[lsearch $objsclass ::WidgetContainer]>=0} {
	$obj foreach_elem [self] update_default
    }
}
FrameVisitor instproc undo_change obj {
    set objsclass [[$obj info class] info heritage]
    #puts "update $obj $objsclass"
    if {[lsearch $objsclass ::InputWidget]>=0} then {	
	$obj setdefault
    } elseif {[lsearch $objsclass ::WidgetContainer]>=0} {
	$obj foreach_elem [self] undo_change
    }
}

Class FrameContainer -superclass {WidgetContainer DoubleList} -parameter {label border}
FrameContainer metadata add attributes
FrameContainer metadata attributes {label border}
Class FrameElement -parameter {
    {side top}
    {anchor c}
    {fill none}
    {expand 0}
}
FrameElement instproc getpackoption {} {
    [self] instvar side anchor fill expand
    return "-side $side -anchor $anchor -fill $fill -expand $expand"
}
# owidget mu eine genestate Klasse von FrameContainer sein
FrameContainer instproc addwidget {owidget paramlist} {
    eval FrameElement ${owidget}::packoption $paramlist
    [self] push_back $owidget
}
FrameContainer instproc map window {
    [self] instvar win border label
    set win $window
    set param {}
    if {[info exist border] && $border=="1"} {
	lappend param -relief solid -borderwidth 1
    }
    if {[info exist label] && $label!=""} {
	# tixLabelFrame hat immer einen Rahmen also param nicht beachten
	tixLabelFrame $window -label $label
	set window [$window subwidget frame]
    } else {
	eval frame $window $param
    }
    # Map Alle Windows
    [self] begin_iterator [self]::it
    while {[[self]::it get]!="NIL"} {
	set elem [[self]::it next_elem]
	regexp {.*::(.+)} $elem _ name
	$elem map ${window}.${name}
    }
    [self]::it destroy
    # Setzt die Widget in die Frame
    [self] begin_iterator [self]::it
    while {[[self]::it get]!="NIL"} {
	set elem [[self]::it next_elem]
	regexp {.*::(.+)} $elem _ name
	eval pack ${window}.${name} [${elem}::packoption getpackoption]
    }
    [self]::it destroy
}
Class NoteBookContainer -superclass {WidgetContainer DoubleList}
# card ist eine Liste {name label}
# OWidget mu eine Subclass des NoteBookContainer sein
# !! Vorsicht die name der Kartei wird auf Kleinschreibung umbewandelt
NoteBookContainer instproc addcard {owidget card} {
    $owidget set cardattr $card
    [self] push_back $owidget
}
NoteBookContainer instproc map {window} {
    [self] instvar win
    set win $window
    tixNoteBook $win
    # Map Alle Windows
    [self] begin_iterator [self]::it
    while {[[self]::it get]!="NIL"} {
	set elem [[self]::it next_elem]
	regexp {.*::(.+)} $elem _ name
	# Auf Kleinschreibung Umstellen
	set name [string tolower $name]
	$elem map ${window}.${name}
    }
    [self]::it destroy
    # Setzt die Widget in die Frame
    [self] begin_iterator [self]::it
    while {[[self]::it get]!="NIL"} {
	set elem [[self]::it next_elem]
	set cardname [lindex [$elem set cardattr] 0]
	set cardname [string tolower $cardname]
	set cardlabel [lindex [$elem set cardattr] 1]
	regexp {.*::(.+)} $elem _ name
	# Auf Kleinschreibung Umstellen
	set name [string tolower $name]
	$win add $cardname -label $cardlabel
	set f [$win subwidget $cardname]
	pack ${window}.${name} -in $f
    }
    [self]::it destroy
    set win $window
}
NoteBookContainer instproc raise {card} {
    [self] instvar win
    $win raise $card
}

# WidgetFactory
# Eine zentralle Stelle fr Widget Erzuegung aus listen XML und anderen
Object WidgetFactory
# Erzeugt imput widget aus listen
# {{typ name label default} ...}
WidgetFactory proc make_simply_inputs {name list} {
    FrameContainer $name
    foreach widget $list {
	set type [lindex $widget 0]
	set wname [lindex $widget 1]
	set label [lindex $widget 2]
	set wdefault [lindex $widget 3]
	switch $type {
	    text {
		OLabelEntry ${name}::$wname -label $label -default $wdefault \
		    -textwidth 20
	    }
	    integer {
		OLabelEntry ${name}::$wname -label $label -default $wdefault
	    }
	    password {
		OLabelEntry ${name}::$wname -label $label -default $wdefault \
		    -show * -textwidth 20
	    }
	    list {
		OComboBox ${name}::$wname -label $label -default $wdefault \
		    -valuelist [lindex $widget 4] 
	    }
	    default {
		error "unsupported widget type $type"
	    }
	}
	$name addwidget ${name}::${wname} {-anchor e}
    }
}
# Idee: list beinhaltet eine Liste mit keydlist
# type und name werden benutzt um Widget zu bilden der Rest
# wird zu optionnen verndert und weiter gegeben
# name und type sind obligatorisch
# z.B.
# {{name data} {type OlabelEntry} {textwidth 10} {label Datenherkunft}}
# packoption werden fr den Packer benutzt wenn vorhanden
WidgetFactory proc make_inputs_keylist {name list} {
    FrameContainer $name
    foreach widget $list {
	set type [keylget widget type]
	set wname [keylget widget name]
	set opt {}
	set lkeys [keylkeys widget]
	foreach key [minusMenge $lkeys {type name packoption}] {
	    lappend opt -$key [keylget widget $key]
	}
	# in einer der nchsten Zeile wird ein Widgetobject gebildet
	eval $type ${name}::$wname $opt
	if {[lcontain $lkeys packoption]} {
	    $name addwidget ${name}::${wname} [keylget widget packoption]
	} else {
	    $name addwidget ${name}::${wname} {-anchor e}
	}
    }
    return $name
}
